type op = Add | Sub | Mult | Div | Mod | Equal | Neq | Less | Leq | Greater | Geq | And | Or
type uop = Neg | Not
type up = Plus | Minus
type typ = Int | Float | Bool | Cry | String | Char | IntList of int | FloatList of int | StringList of int
type bind = typ * string

type expr =
    Num of int
  | Blit of bool
  | Var of string
  | Chara of char
  | Dec of float
  | Slit of string
  | Binop of expr * op * expr
  | Unop of uop * expr
  | Unp of expr * up
  | Assign of string * expr
  | Call of string * expr list
  | IntListLit of expr list
  | FloatListLit of expr list
  | StringListLit of expr list
  | ArrayAccess of string * expr


type stmt =
   Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | While of expr * stmt
  | Vdecl of typ * string
  | VdeclAssign of typ * string * expr

type func_decl = {
  typ: typ;
  fname: string;
  formals: bind list;
  body: stmt list;
}

type program =  func_decl list
let string_of_op = function
  Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "and"
  | Or -> "or"
  | Mod -> "%"

  let string_of_uop = function
  Neg -> "-"
| Not -> "not"

let rec string_of_expr = function
  Call(f, el) -> f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ");\n"
  | Slit l -> l
  | Blit(true) -> "true"
  | Blit(false) ->  "false"
  | Num n -> string_of_int n
  | Dec d -> string_of_float d
  | Chara c -> String.make 1 c
  | Var v -> v
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | Unop(uop, e) -> string_of_uop uop ^ " " ^ string_of_expr e
  | Binop(v, op, e) -> string_of_expr v ^ " " ^ string_of_op op ^ " " ^ string_of_expr e
  | IntListLit(elements) -> "[" ^ String.concat ", " (List.map string_of_expr elements) ^ "]"
  | FloatListLit(elements) -> "[" ^ String.concat ", " (List.map string_of_expr elements) ^ "]"
  | StringListLit(elements) -> "[" ^ String.concat ", " (List.map string_of_expr elements) ^ "]"
  | ArrayAccess(n, index) -> n ^ "[" ^ string_of_expr index ^ "]"
  | _ -> ""

let string_of_typ = function
  Cry -> "cry"
  | Int -> "int"
  | Float -> "float"
  | Bool -> "cry"
  | String -> "string"
  | Char -> "char"
  | IntList x -> "array<int, " ^ (string_of_int x) ^ ">"
  | FloatList x -> "array<float, " ^ (string_of_int x) ^ ">"
  | StringList x -> "array<str, " ^ (string_of_int x) ^ ">"

let rec string_of_stmt = function
   | Block(stmts) ->
    "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n"
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s
  | Vdecl(t, s) -> string_of_typ t ^ " " ^ s ^ ";\n"
  | VdeclAssign(t, s, e) -> string_of_typ t ^ " " ^ s ^ " = " ^ string_of_expr e ^ ";\n"

let string_of_func_decl func_decl =
    string_of_typ func_decl.typ ^ " " ^
    func_decl.fname ^ "(" ^ String.concat ", " (List.map snd func_decl.formals) ^
    ")\n{\n" ^
    String.concat "" (List.map string_of_stmt func_decl.body) ^
    "}\n"

let string_of_program (funcs) =
  String.concat "\n" (List.map string_of_func_decl funcs)
