(* Semantically-checked Abstract Syntax Tree and functions for printing it *)

open Ast

type sexpr = typ * sx
and sx =
    SNum of int
  | SBlit of bool
  | SVar of string
  | SChara of char
  | SDec of float
  | SSlit of string
  | SBinop of sexpr * op * sexpr
  | SUnop of uop * sexpr
  | SAssign of string * sexpr
  | SCall of string * sexpr list
  | SIntListLit of sexpr list
  | SFloatListLit of sexpr list
  | SStringListLit of sexpr list
  | SArrayAccess of string * sexpr

type sstmt =
    SBlock of sstmt list
  | SExpr of sexpr
  | SReturn of sexpr
  | SIf of sexpr * sstmt * sstmt
  | SFor of sexpr * sexpr * sexpr * sstmt
  | SWhile of sexpr * sstmt
  | SVdecl of typ * string
  | SVdeclAssign of typ * string * sexpr

type sfunc_decl = {
    styp : typ;
    sfname : string;
    sformals : bind list;
    sbody : sstmt list;
  }

type sprogram = sfunc_decl list

(* Pretty-printing functions, used when printing out error messages *)
 let rec string_of_sexpr (t, e) =
  "(" ^ string_of_typ t ^ " : " ^ (match e with
    SSlit(l) -> l
  | SNum(n) -> string_of_int n
  | SDec(d) -> string_of_float d
  | SVar(v) -> v
  | SAssign(v, e) -> v ^ " = " ^ string_of_sexpr e
  | SBinop(e1, op, e2) -> string_of_sexpr e1 ^ " " ^ string_of_op op ^ " " ^ string_of_sexpr e2

  | SCall(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_sexpr el) ^ ")"
  | SIntListLit(el) -> "[" ^ String.concat ", " (List.map string_of_sexpr el) ^ "]"
  | SFloatListLit(el) -> "[" ^ String.concat ", " (List.map string_of_sexpr el) ^ "]"
  | SStringListLit(el) -> "[" ^ String.concat ", " (List.map string_of_sexpr el) ^ "]"
  | SArrayAccess(name,e) -> name ^ "[" ^ string_of_sexpr e^ "]"
  | _ -> "") ^ ")"

let rec string_of_sstmt = function
    SBlock(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_sstmt stmts) ^ "}\n"
  | SExpr(expr) -> string_of_sexpr expr ^ ";\n";
  | SVdeclAssign(t, s, e) -> string_of_typ t ^ " " ^ s ^ " = " ^ string_of_sexpr e ^ ";\n"
  | SVdecl(t, s) -> string_of_typ t ^ " " ^ s ^ ";\n"
  | _ -> ""


let string_of_sfdecl fdecl =
  string_of_typ fdecl.styp ^ " " ^
  fdecl.sfname ^ "(" ^ String.concat ", " (List.map snd fdecl.sformals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_sstmt fdecl.sbody) ^
  "}\n"

let string_of_sprogram funcs =
  String.concat "\n" (List.map string_of_sfdecl funcs)
