
open Ast
open Sast

module StringMap = Map.Make(String)


let check (functions) =

  let built_in_decls =
    let add_bind map (name, ty) = StringMap.add name {
      typ = Cry;
      fname = name;
      formals = [(ty, "x")];
      body = [] } map
    in List.fold_left add_bind StringMap.empty [(("print", String))]
  in

  (* Add function names to map and catch duplicates *)
  let add_func map fd =
    let built_in_err = "the function " ^ fd.fname ^ " is built-in and may not be redefined" 
    and dup_err = "duplicate function: a function with name " ^ fd.fname ^ " has already been defined"
    and make_err er = raise (Failure er)
    and n = fd.fname (* Name of the function *)
    in match fd with (* check for duplicate functions or redefinition of built-in functions *)
       _ when StringMap.mem n built_in_decls -> make_err built_in_err
       | _ when StringMap.mem n map -> make_err dup_err  
       | _ ->  StringMap.add n fd map 
  in

  let function_decls = List.fold_left add_func built_in_decls functions
  (* Add all built in functions to this table, including all 3 types of print*)
  in

  let find_func s =
    if String.equal s "main" then
      (try StringMap.find s function_decls
      with Not_found -> raise (Failure ("missing main function")))
    else 
      try StringMap.find s function_decls
      with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in
  let _= find_func "main" in

  let check_function m_locals func =
    let check_assign funcname lvaluet rvaluet err =
      (* Special case for print function that takes any input *)
      if funcname = "print" ||
      lvaluet = rvaluet then lvaluet
      else raise (Failure err)
    in

    (* check_assign for var types *)
    let check_assign_vars lvaluet rvaluet err = match (lvaluet, rvaluet) with
      (Int, IntList(_)) -> lvaluet
      | (Float, FloatList(_)) -> lvaluet
      | (String, StringList(_)) -> lvaluet
      | _ -> if lvaluet = rvaluet then lvaluet else raise (Failure err)
    in

    (* Build local vars table for this function *)
    let locals = List.fold_left (fun m (typ, name) ->
      StringMap.add name typ m)
        m_locals func.formals
    in

    (* Return a variable from our local symbol table *)
    let type_of_identifier s m =
      try StringMap.find s m
      with Not_found -> raise (Failure ("undeclared identifier " ^ s))
    in

    let rec expr m_locals = function
     Slit l    -> (String, SSlit l)
     | Num i   -> (Int, SNum i)
     | Dec d   -> (Float, SDec d)
     | Chara c  -> (Char, SChara c)
     | Blit b  -> (Bool, SBlit b)
     | Var v   -> (type_of_identifier v m_locals, SVar v)
     | IntListLit values ->
      let length = List.length values in
      let l' = List.map (expr m_locals) values in
      (match l' with
        (_,_) :: _ -> (IntList(length), SIntListLit(l'))
        |[] -> (IntList(0), SIntListLit(l')))
     | FloatListLit values ->
      let length = List.length values in
      let l' = List.map (expr m_locals) values in
      (match l' with
        (_, _) :: _ -> (FloatList(length), SFloatListLit(l'))
        | [] -> (FloatList(0), SFloatListLit(l')))
     | StringListLit values ->
        let length = List.length values in
        let l' = List.map (expr m_locals) values in
        (match l' with
          (_, _) :: _ -> (StringList(length), SStringListLit(l'))
          | [] -> (StringList(0), SStringListLit(l')))
     | ArrayAccess(name, index) ->
     let leftside = (type_of_identifier name m_locals)
     in let v = expr m_locals index in (leftside, SArrayAccess(name, v))
     | Unp (_,_) -> raise (Failure ("error: this operator is unsupported"))
     | Assign(var, e) as ex ->
       let lt = type_of_identifier var m_locals
       and (rt, e') = expr m_locals e in
       let err = "illegal assignment " ^ string_of_typ lt ^ " = " ^
            string_of_typ rt ^ " in " ^ string_of_expr ex
          in (check_assign_vars lt rt err, SAssign(var, (rt, e')))
     | Binop(e1, op, e2) as e ->
      let (t1, e1') = expr m_locals e1
      and (t2, e2') = expr m_locals e2 in
      (* All binary operators require operands of the same type *)
       let same = t1 = t2 in
      (* Determine expression type based on operator and operand types *)
       let ty = match op with
         Add | Sub | Mult | Div | Mod when same && t1 = Int   -> Int
        | Add | Sub | Mult | Div when same && t1 = Float -> Float
        | Equal | Neq            when same               -> Bool
        | Less | Leq | Greater | Geq
                when same && (t1 = Int || t1 = Float) -> Bool
        | And | Or when same && t1 = Bool -> Bool
        | _ -> raise (Failure ("illegal binary operator " ^
                  string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
                  string_of_typ t2 ^ " in " ^ string_of_expr e))
      in (ty, SBinop((t1, e1'), op, (t2, e2')))
     | Call(fname, args) as call -> (* Now dealing with every function call *)
        let fd = find_func fname in
        let param_length = List.length fd.formals in
        if List.length args != param_length then
         raise (Failure ("expecting " ^ string_of_int param_length ^
                        " arguments in " ^ string_of_expr call))
        else let check_call (ft, _) e =
          let (et, e') = expr m_locals e in
          let err = " illegal argument found " ^ string_of_typ et ^
            " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e
          in (check_assign fname ft et err, e')
        in
        let args' = List.map2 check_call fd.formals args
        in
        (fd.typ, SCall(fname, args'))
      | Unop(op, e) as ex ->
        let (t, e') = expr m_locals e in
        let ty = match op with
          Neg when t = Int || t = Float -> t
        | Not when t = Bool -> Bool
        | _ -> raise (Failure ("illegal unary operator " ^
                                string_of_uop op ^ " " ^ string_of_typ t ^
                                " in " ^ string_of_expr ex))
        in (ty, SUnop(op, (t, e')))
    in
    let check_bool_expr m_locals e =
      let (t', e') = expr m_locals e
      and err = "expected Boolean expression, got " ^ string_of_expr e ^ " instead"
      in if t' != Bool then raise (Failure err) else (t', e')
    in
          (* Return type: tuple --> (StringMap, Sast.sstmt) *)
    let rec check_stmt m_locals = function
      Expr e -> (m_locals, SExpr (expr m_locals e))
      | If(p, b1, b2) -> let (_, b1') = check_stmt m_locals b1 and
                        (_, b2') = check_stmt m_locals b2 in
        (m_locals, SIf(check_bool_expr m_locals p, b1', b2'))
      | While (p, b1) -> let (_, b1') = check_stmt m_locals b1 in
                        (m_locals, SWhile(check_bool_expr m_locals p, b1'))
      | For (e1,e2,e3,b1) -> let (_, b1') = check_stmt m_locals b1 in
                        (m_locals, SFor(expr m_locals e1, check_bool_expr m_locals e2,
                        expr m_locals e3, b1'))
      | Return e -> let (t, e') = expr m_locals e in
        if t = func.typ then (m_locals, SReturn (t, e'))
        else raise ( Failure ("return gives " ^ string_of_typ t ^ " expected " ^
            string_of_typ func.typ ^ " in " ^ string_of_expr e))
      | Vdecl(t, var) -> (StringMap.add var t m_locals, SVdecl (t, var))
      | VdeclAssign (t, var, e) as st ->
        let (rt, e') = expr m_locals e in
        let err = "illegal assignment " ^ string_of_typ t ^ " = " ^
            string_of_typ rt ^ " in " ^ string_of_stmt st in
            let m_locals = StringMap.add var t m_locals in
            (m_locals, SVdeclAssign((check_assign_vars t rt err), var, (rt, e')))
      | Block sl ->
        (* Return type: tuple --> (StringMap, Sast.sstmt list) *)
        let rec check_stmt_list m_locals = function
            [Return _ as s] -> let (m_locals', st) = check_stmt m_locals s
                               in (m_locals', [st])
          | Return _ :: _   -> raise (Failure "nothing may follow a return")
          | Block sl :: ss  -> check_stmt_list m_locals (sl @ ss) (* Flatten blocks *)
          | s :: ss         -> let (ml, st) = check_stmt m_locals s in (* new m_locals, translated instruction, change format*)
                               let (ml', st') = check_stmt_list ml ss in
                               let new_list = st :: st' in
                               (ml', new_list)
          | []              -> (m_locals, [])
        in let (m_locals', sl') = check_stmt_list m_locals sl
        in (m_locals', SBlock(sl'))

    in

    (* body of check_function *)
    { styp = func.typ;
        sfname = func.fname;
        sformals = func.formals;
        sbody = let (_, st') = check_stmt locals (Block func.body) in
        match st' with
        SBlock(sl) -> sl
        | _ -> raise (Failure ("internal error: block didn't become a block?"))
    }
  in (List.map (check_function StringMap.empty) functions);
